Relog = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0","AceDB-2.0","AceEvent-2.0","Parser-3.0")

local ConsoleOptions = {
	type='group',
	args={
		Show={
			name = "Show",
			desc = "Shows the table window",
			type = 'execute',
			func = function() Relog.TableWindow:Show() end
		},
		Reset={
			name = "Reset",
			desc = "Resets the data",
			type = 'execute',
			func = function() Relog:ResetData() end
		},
	}
}

Relog:RegisterChatCommand({"/relog"}, ConsoleOptions)

local DefaultConfig={
	--Abbreviations reduce space stored for the log
	Abbrev={},	--Table converts from string to abbreviation (a number)
	RAbbrev={},	--Table converts from abbreviation to string

	--The parsed combat log
	LogOffset=0,
	StartLog=1,
	LogSize=0,
	NextLog=1,
	Log={
		Time={},
		Source={},
		Victim={},
		Type={},	--What type of log message is this? ("Damage" is most common so default)
		Ability={},	--false is most common so set as default
		Result={},	--"Hit" is most common so used as default
		Amount={},

		--Special Data for Dispel
		VictimAbility={},

		--Special Data for Drain/Gain
		Attribute={},

		--Leech Info
		SourceGained={},
		AttributeGained={},
		AmountGained={},

		--Reputation
		Faction={},
		Rank={},

		--Special Data for Damage
		DamageType={},
		AbsorbAmount={['*']=0},
		BlockAmount={['*']=0},
		ResistAmount={['*']=0},
		VulnerableAmount={['*']=0},

		--Healing
		Overheal={},

		--Item Data
		ItemLink={},
		ItemName={},
		ItemRarity={},
		ItemString={},

		--Items just for Experience which requires several
		BonusAmount={['*']=0},
		BonusType={},
		PenaltyAmount={['*']=0},
		PenaltyType={},
		AmountRaidPenalty={['*']=0},
		AmountGroupBonus={['*']=0},
		CompiledCache={},
	},

	--Units and their time based data
	Units={['*']={
			LastUpdated=0,
			NumLogs=0,
			
			--Attributes
			UnitID={},
			UnitIDTime={},
			Friendly={},
			FriendlyTime={},
			isPlayer={},
			isPlayerTime={},
			isGrouped={},
			isGroupedTime={},
			lClass={},
			lClassTime={},
			Class={},
			ClassTime={},
			Faction={},
			FactionTime={},
			Level={},
			LevelTime={},
			
			--Stats
			Health={},
			HealthTime={},
			HealthMax={},
			HealthMaxTime={},
			Power={},
			PowerTime={},
			PowerMax={},
			PowerMaxTime={},
			PowerType={},
			PowerTypeTime={},
			MinRange={},
			MinRangeTime={},
			MaxRange={},
			MaxRangeTime={},

			--PvP (use with GetPVPRankInfo and UnitPVPRank)
			RankName={},
			RankNameTime={},
			RankNumber={},
			RankNumberTime={},
		}
	}
}

local ProfileDefaults={
	MaxLogs=5000,	
	RowHeight=12,
	Font="Arial Narrow",
	OutputLog=false,
}

Relog:RegisterDB("RelogDB","RelogPerCharDB")
Relog:RegisterDefaults('char',DefaultConfig)	
Relog:RegisterDefaults('profile',ProfileDefaults)	




--Subsecond determining by enclosing a range at second flips
Relog.SubLower=0
Relog.SubWidth=2

local AcceptedWidth=0.0005

function Relog:DetermineSubsecond()
	local UpperTime=GetTime()
	local CurTime=time()

	if (CurTime-Relog.SubLastTime2)==1 then
		--1 second difference
		local LowerTime=Relog.SubLastTime
		local Width=UpperTime-LowerTime
		LowerTime=LowerTime-math.floor(LowerTime)

		if Relog.SubLower<LowerTime then
			Relog.SubWidth=Relog.SubWidth-(LowerTime-Relog.SubLower)
			Relog.SubLower=LowerTime

			--Relog:Print("Raised Lower to "..LowerTime)
		end

		if (Relog.SubWidth+Relog.SubLower)>(Width+LowerTime) then
			Relog.SubWidth=Width+LowerTime-Relog.SubLower
			--Relog:Print("Reduced Width to "..Relog.SubWidth)
		end

		if Relog.SubWidth<AcceptedWidth then
			--Relog:Print("Finished with range of ".. Relog.SubLower .." to ".. Relog.SubLower+Relog.SubWidth)
			Relog:CancelScheduledEvent("DetermineSubSecond")
		end
	end
	
	Relog.SubLastTime=UpperTime
	Relog.SubLastTime2=CurTime
end

function Relog:GetSubsecond()
	local T=GetTime()
	T=T-Relog.SubLower+Relog.SubWidth/2
	T=T-math.floor(T)
	return T
end

function Relog:GetTime()
	return time()+Relog:GetSubsecond()
end



function Relog:OnInitialize()
	Relog.SubLastTime=GetTime()
	Relog.SubLastTime2=time()

	Relog:ScheduleRepeatingEvent("DetermineSubSecond",Relog.DetermineSubsecond,0)
	Relog:CreateTableWindow()
end

function Relog:OnEnable(first)
	Relog.Parse.SetupParser()
	--Relog:ResetData()
end

function Relog:ResetData()
	for _,v in pairs(Relog.db.char.Log) do
		for k in pairs(v) do
			v[k]=nil
		end
	end
	Relog.db.char.StartLog=1
	Relog.db.char.NextLog=1
	Relog.db.char.LogSize=0

	for k in pairs(Relog.db.char.Units) do
		Relog.db.char.Units[k]=nil
	end
	Relog.db.char.Abrrev={}
	Relog.db.char.RAbrrev={}
	Relog:ResetDB("char")
	Relog.Window:RefreshRows()
	Relog:ResetFilterAndSort()
	Relog.TableWindow.Table:UpdateRows()
end


function Relog:ChangeMaxLogs(NewVal)
	local OldVal=Relog.db.profile.MaxLogs
	Relog.db.profile.MaxLogs=NewVal

	local Start=Relog.db.char.StartLog
	local Size=Relog.db.char.LogSize

	if NewVal>OldVal and (Start+Size)>=OldVal then
		local NewPos=NewVal
		for i=OldVal,Start,-1 do
			for k,v in pairs(Relog.db.char.Log) do 
				v[NewPos]=v[i]
				v[i]=nil
			end
			NewPos=NewPos-1
		end
		Relog.db.char.StartLog=Relog.db.char.StartLog+NewVal-OldVal
	elseif NewVal<OldVal then
		local NewStart=Start-(OldVal-NewVal)

		if NewStart<1 then
			NewStart=NewStart+NewVal
		end

		Relog.db.char.StartLog=NewStart

		local NewPos=NewStart
		local OldPos=Start
		
		if Relog.db.char.LogSize>NewVal then
			OldPos=OldPos+(Relog.db.char.LogSize-NewVal)
		end
		if OldPos>OldVal then
			OldPos=OldPos-OldVal
		end

		for i=1,NewVal do
			for k, v in pairs(Relog.db.char.Log) do 
				v[NewPos]=v[OldPos]
				v[OldPos]=nil
			end
			NewPos=NewPos+1
			OldPos=OldPos+1
			if NewPos>NewVal then
				NewPos=NewPos-NewVal
			end
			if OldPos>OldVal then
				OldPos=OldPos-OldVal
			end
		end

		for i=NewVal+1,OldVal do
			for k,v in pairs(Relog.db.char.Log) do 
				v[i]=nil
			end
		end

		if Relog.db.char.LogSize>NewVal then
			Relog.db.char.LogSize=NewVal
		end
	end
end


local Tables={}
function Relog:FreeTable(t)
	if type(t)~="table" then
		return
	end

	for k in pairs(t) do
		t[k]=nil
	end

	for _,v in pairs(Tables) do
		if v==t then
			return
		end
	end

	tinsert(Tables,t)
end

function Relog:FreeTableRecurse(t)
	--Check the table first before recursing
	for _,v in pairs(Tables) do
		if v==t then
			return
		end
	end

	for k in pairs(t) do
		if type(t[k])=="table" then
			Relog:FreeTableRecurse(t[k])
		end
		t[k]=nil
	end
	
	tinsert(Tables,t)
end

function Relog:FreeTableRecurseLimit(t,depth)
	--Check the table first before recursing
	if depth<0 then
		return
	end

	for k in pairs(t) do
		if type(t[k])=="table" then
			Relog:FreeTableRecurseLimit(t[k],depth-1)
		end
		t[k]=nil
	end
	
	tinsert(Tables,t)
end

function Relog:GetTable()	
	local t
	if #Tables>0 then
		t=Tables[1]
		tremove(Tables,1)
		if #t>0 then
			Relog:Print("WARNING! For some reason there is "..#t.." entries left. There is probably a table in use that shouldn't have been freed")
		end
		return t
	else
		return {}
	end
end

function Relog:HowManyTables(str)
	if str==nil then
		str=""
	else
		str=str.." "
	end
	Relog:Print(str.."Free Tables: "..#Tables)
end

function Relog:ResetTableCache()
	Tables=Relog:GetTable()
end